#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <pwd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <unistd.h>
#include <stdint.h>
#include "lve-ctl.h"

/**
 * SecureLVE definitions
 */
static const int  BUFSIZE = 8192;
static const char MOUNT_TEST_STRING[]={"%s/%s/proc/cpuinfo"};
static const char USER_TMP_DIR[]={"%s/.securelve-tmp"};
static const int  USER_TMP_MODE_T = 0700;


static int 
print_error(char *str, int err, const char *format,...)
{
	va_list         ap;
	int             result;

	if (str == NULL)
		return 0;

	va_start(ap, format);
	result  = vsnprintf(str, BUFSIZE, format, ap);
	va_end(ap);
	errno = err;

	return result;
}

static int 
mount_is_error(int rc)
{
	return rc != 0 && rc != EBUSY;
}

int mkdir_parents(const char *path, mode_t mode)
{
	const char *p;
	if (mkdir(path, mode) < 0 && errno != EEXIST) {
		if (errno != ENOENT)
			return -1;

		p = strrchr(path, '/');
		if (p == NULL || p == path)
			return -1; /* shouldn't happen */

		size_t size = (size_t) ((const char*) p - (const char *) path);
		char *tmp = calloc(sizeof(char), size + 1);
		memcpy(tmp, path, size);
		if (mkdir_parents(tmp, mode) < 0) {
			free(tmp);
			return -1;
		}
		free(tmp);

		/* should work now */
		if (mkdir(path, mode) < 0 && errno != EEXIST)
			return -1;
	}

	return 0;
}

void trim(char *str) {
  int len = strlen(str);
  int i;
  for (i = len - 1; i>=0; i++) {
	if (str[i] == ' ' || str[i] == '\n' || str[i] == '\r') { 
	  str[i]=0; 
        } else return;
  }
}

int 
mount_dir(const char *jaildir, const char *subdir, const char *basedir, char *error_str)
{
	char            src[BUFSIZE];
	char            target[BUFSIZE];
	char            line[BUFSIZE];
	int		rc;
	FILE           *file;
        struct stat	st;

	file = fopen(SECURELVE_MP_FILE, "r");
	if (file == NULL) {
	  print_error(error_str, errno, "Unable to open config file %s:%s", SECURELVE_MP_FILE, strerror(errno));
	  return 0;
	}
	while (fgets(line, BUFSIZE, file) != NULL) {
	  trim(line);
	  if (line[0] == 0 || line[0] == '#') continue; //skip
	  if (line[0] == '/') {
	    snprintf(src, BUFSIZE, "%s", line);
	    snprintf(target, BUFSIZE, "%s/%s%s", basedir, subdir, line);
	  } else {
	    snprintf(src, BUFSIZE, "%s/%s", jaildir, line);
	    snprintf(target, BUFSIZE, "%s/%s/%s", basedir, subdir, line);
	  }
          if (stat(src, &st) == -1) {
                continue;
          }
	  mkdir_parents(target, 0755);
	  rc = mount(src, target, NULL, MS_MGC_VAL | MS_BIND | MS_RDONLY , NULL); // MS_NOSUID flag should be added
	  if (mount_is_error(rc)) {
	    print_error(error_str, errno, "Unable to mount %s -> %s: %s", src, target, strerror(errno));
	    return 0;
	  }
	}
	fclose(file);
	return 1;

}


int 
mount_jail(const char *homedir, const char *username, const char *basedir, const char *jaildir, char *error_str,
	   uid_t uid, gid_t gid)
{
	char            src[BUFSIZE];
	char            target[BUFSIZE];
	struct stat     st;

	/* test if mounted */
	snprintf(target, BUFSIZE, MOUNT_TEST_STRING, basedir, username);
	if (stat(target, &st) == 0) {
		return 1;
	}
	if (!mount_dir(jaildir, username, basedir, error_str))
		return 0;

	/* mount home directory */
	snprintf(target, BUFSIZE, "%s/%s%s", basedir, username, homedir);
	if (mount_is_error(mount(homedir, target, NULL, MS_MGC_VAL | MS_BIND | MS_NOSUID, NULL))) {
		print_error(error_str, errno, "Unable to mount %s -> %s: %s", homedir, target, strerror(errno));
		return 0;
	}

	/* mount tmp directory */
	snprintf(src, BUFSIZE, USER_TMP_DIR, homedir);
	snprintf(target, BUFSIZE, "%s/%s/tmp", basedir, username);
	if (stat(src, &st) == -1) {
		if (errno == ENOENT) {
			if (mkdir(src, USER_TMP_MODE_T) != 0) {
				if (errno != EEXIST) {
					print_error(error_str, errno, "Unable to create %s", src, strerror(errno));
					return 0;
				}
			}
			if (chown(src, uid, gid) != 0) {
				print_error(error_str, errno, "Unable to set correct ownership to %s", src, strerror(errno));
				return 0;
			}
		} else {
			print_error(error_str, errno, "Unable to stat %s", src, strerror(errno));
		}
	}
	if (mount_is_error(mount(src, target, NULL, MS_MGC_VAL | MS_BIND | MS_NOSUID, NULL))) {
		print_error(error_str, errno, "Unable to mount %s -> %s: %s", src, target, strerror(errno));
		return 0;
	}
	/* mount /proc directory */
	snprintf(target, BUFSIZE, "%s/%s/proc", basedir, username);
	if (mount_is_error(mount("/proc", target, NULL, MS_MGC_VAL | MS_BIND | MS_NOSUID, NULL))) {
		print_error(error_str, errno, "Unable to mount %s -> %s: %s", homedir, target, strerror(errno));
		return 0;
	}
	return 1;
}

/* 
   int umount_jail(const char *basedir, const char *username, char *error_str) {
   char target[BUFSIZE];
   char line[BUFSIZE];
   FILE *in;

   snprintf(target, BUFSIZE, "%s/%s", basedir, username);
   in = fopen("/proc/mounts", "r");
   if (in == NULL) {
   print_error(error_str, errno, "Unable to open /proc/mounts unmounting user %s: %s", username, strerror(errno));    
   return 0;
   }
   while (fgets(line, BUFSIZE, in)) {
   char *c, *e = line;
   for (c = line; *c; c++) {
   if (*c == ' ') {
   e = c+1;
   break;
   }
   }

   for (c = e; *c; c++) {
   if (*c == ' ') {
   *c = '\0';
   break;
   }
   }
   if (strstr(target, e) == e) {
   umount2(e, MNT_DETACH);
   }
   }
   fclose(in);
   return 1;
   }
*/

int 
jail(struct passwd * pw, char *error_str)
{
	char            jaildir[BUFSIZE], cwd[BUFSIZE];

	if (strcmp(pw->pw_shell, SECURELVE_SHELL) != 0) {
		return 0;
	}

	/* init current working dir */
	if (getcwd(cwd, BUFSIZE) == NULL) {
		strcpy(cwd, "/");
	}
	snprintf(jaildir, BUFSIZE, "%s/%s", SECURELVE_BASEDIR, pw->pw_name);

	if (mount_jail(pw->pw_dir, pw->pw_name, SECURELVE_BASEDIR, SECURELVE_JAIL, error_str, pw->pw_uid, pw->pw_gid) == 1) {
		if (chroot(jaildir)) {
			print_error(error_str, errno, "abort, chroot(%s) failed: %s, check the permissions for %s", jaildir,
				    strerror(errno), jaildir);
			return -1;
		}
		/* success */
		int rc = chdir(cwd);
		return 1;
	}
	/*
	 * } else { umount_jail(SECURELVE_BASEDIR, pw->pw_name, NULL); }
	 */
	/* error */
	return -1;
}

int 
lve_jail(struct passwd * pw, char *error_str)
{
	return jail(pw, error_str);
}
